use anyhow::Context as _;
use std::collections::BTreeMap;

/// Read all issues to generate the full techtree
pub async fn collect_tree(
    forgejo: &forgejo_api::Forgejo,
    meta: &crate::event_meta::IssueEventMeta,
) -> anyhow::Result<crate::tree::Tree> {
    let issues = forgejo
        .issue_list_issues(
            &meta.issue.repository.owner,
            &meta.issue.repository.name,
            forgejo_api::structs::IssueListIssuesQuery {
                // We also want the closed issues
                state: Some(forgejo_api::structs::IssueListIssuesQueryState::All),
                // No pagination
                limit: None,
                // Only issues
                r#type: Some(forgejo_api::structs::IssueListIssuesQueryType::Issues),
                ..Default::default()
            },
        )
        .await
        .context("Failed fetching issue list")?;

    let mut tree = crate::tree::Tree::new();
    let mut issue_numbers = Vec::new();

    for issue in issues.iter() {
        let element = match element_from_issue(issue) {
            Ok(el) => el,
            Err(e) => {
                let maybe_number = issue
                    .number
                    .map(|n| n.to_string())
                    .unwrap_or("<unknown>".to_owned());
                log::warn!("Failed processing issue #{maybe_number}: {e:?}");
                continue;
            }
        };

        issue_numbers.push(element.issue_number);
        tree.add_element(element);
    }

    for issue in issue_numbers.into_iter() {
        let dependencies = forgejo
            .issue_list_issue_dependencies(
                &meta.issue.repository.owner,
                &meta.issue.repository.name,
                &issue.to_string(),
                forgejo_api::structs::IssueListIssueDependenciesQuery {
                    limit: None,
                    ..Default::default()
                },
            )
            .await
            .with_context(|| {
                format!(
                    "Failed to fetch issue dependencies for #{}",
                    meta.issue.number
                )
            })?;

        for dep in dependencies {
            let dep_number = dep.number.context("Missing issue number in dependency")?;
            if !tree.find_element_by_issue_number(dep_number).is_some() {
                log::warn!("Found dependency from #{issue} on non-tracked issue #{dep_number}!");
            }

            tree.add_dependency_by_issue_number(issue, dep_number);
        }
    }

    Ok(tree)
}

fn element_from_issue(issue: &forgejo_api::structs::Issue) -> anyhow::Result<crate::tree::Element> {
    let issue_number = issue.number.context("Missing issue number")?;
    let description = issue
        .title
        .as_deref()
        .context("Issue is missing a title")?
        .to_owned();

    let ty_labels: Vec<_> = issue
        .labels
        .as_ref()
        .context("Issue does not have any labels")?
        .iter()
        .filter_map(|l| l.name.as_deref())
        .filter(|l| l.starts_with("ty/"))
        .collect();

    if ty_labels.len() == 0 {
        anyhow::bail!("Issue #{issue_number} has no type label!");
    }
    if ty_labels.len() > 1 {
        anyhow::bail!("Issue #{issue_number} has more than one type label!");
    }

    let ty = match *ty_labels.first().unwrap() {
        "ty/equipment" => crate::tree::ElementType::Equipment,
        "ty/process" => crate::tree::ElementType::Process,
        "ty/knowledge" => crate::tree::ElementType::Knowledge,
        t => anyhow::bail!("Unknown element type for issue #{issue_number}: {t:?}"),
    };

    let status = match issue.state.context("Missing issue state")? {
        forgejo_api::structs::StateType::Open => {
            if issue.assignee.is_some()
                || issue
                    .assignees
                    .as_ref()
                    .map(|v| v.len() > 0)
                    .unwrap_or(false)
            {
                crate::tree::ElementStatus::Assigned
            } else {
                crate::tree::ElementStatus::Missing
            }
        }
        forgejo_api::structs::StateType::Closed => crate::tree::ElementStatus::Completed,
    };

    Ok(crate::tree::Element {
        issue_number,
        description,
        ty,
        status,
    })
}
